/*
  This file is part of CDO. CDO is a collection of Operators to
  manipulate and analyse Climate model Data.

  Copyright (C) 2003-2020 Uwe Schulzweida, <uwe.schulzweida AT mpimet.mpg.de>
  See COPYING file for copying and redistribution conditions.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
*/

/*
   This module contains the following operators:

      Vertint    gh2hl           Model geometric height level to height level interpolation
*/

#include <cdi.h>

#include "cdo_options.h"
#include "process_int.h"
#include "cdo_vlist.h"
#include "field_vinterp.h"
#include "stdnametable.h"
#include "util_string.h"
#include "const.h"
#include "cdo_zaxis.h"
#include "param_conversion.h"


template <typename T>
static double
vert_interp_lev_kernel(double w1, double w2, const T var1L1, const T var1L2, const T missval)
{
  if (DBL_IS_EQUAL(var1L1, missval)) w1 = 0.0;
  if (DBL_IS_EQUAL(var1L2, missval)) w2 = 0.0;

  // clang-format off
  if      (IS_EQUAL(w1, 0) && IS_EQUAL(w2, 0)) return missval;
  else if (IS_EQUAL(w1, 0)) return (w2 >= 0.5) ? var1L2 : missval;
  else if (IS_EQUAL(w2, 0)) return (w1 >= 0.5) ? var1L1 : missval;
  else                      return var1L1 * w1 + var1L2 * w2;
  // clang-format on
}

//  3D vertical interpolation
template <typename T>
static void
vert_interp_lev3d(size_t gridsize, T missval, const Varray<T> &vardata1, Varray<T> &vardata2, int nlev2,
                  const Varray<int> &lev_idx1, const Varray<int> &lev_idx2, const Varray<double> &lev_wgt1, const Varray<double> &lev_wgt2)
{
  for (int ilev = 0; ilev < nlev2; ilev++)
    {
      auto offset = ilev * gridsize;
      auto var2 = &vardata2[offset];

#ifdef _OPENMP
#pragma omp parallel for default(none) shared(gridsize, offset, vardata1, var2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2, missval)
#endif
      for (size_t i = 0; i < gridsize; i++)
        {
          const auto idx1 = lev_idx1[offset + i];
          const auto idx2 = lev_idx2[offset + i];
          const auto wgt1 = lev_wgt1[offset + i];
          const auto wgt2 = lev_wgt2[offset + i];

          // upper/lower values from input field
          const auto var1L1 = vardata1[idx1];
          const auto var1L2 = vardata1[idx2];

          var2[i] = vert_interp_lev_kernel(wgt1, wgt2, var1L1, var1L2, missval);
        }
    }
}

void
vert_interp_lev3d(size_t gridsize, double missval, Field3D &field1, Field3D &field2, int nlev2,
                  const Varray<int> &lev_idx1, const Varray<int> &lev_idx2, const Varray<double> &lev_wgt1, const Varray<double> &lev_wgt2)
{
  if (field1.memType == MemType::Float)
    {
      vert_interp_lev3d(gridsize, (float)missval, field1.vec_f, field2.vec_f, nlev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
  else
    {
      vert_interp_lev3d(gridsize, missval, field1.vec_d, field2.vec_d, nlev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
}

static void
vert_gen_weights(int expol, int nlev1, const Varray<double> &lev1, int nlev2, const Varray<double> &lev2, Varray<int> &lev_idx1, Varray<int> &lev_idx2,
                 Varray<double> &lev_wgt1, Varray<double> &lev_wgt2)
{
  for (int i2 = 0; i2 < nlev2; ++i2)
    {
      int idx1 = 0, idx2 = 0;
      double val1, val2 = 0.0;

      // Because 2 levels were added to the source vertical coordinate (one on top, one at the bottom), its loop starts at 1
      int i1;
      for (i1 = 1; i1 < nlev1; ++i1)
        {
          const auto lev1isless = (lev1[i1 - 1] < lev1[i1]);
          idx1 = lev1isless ? i1 - 1 : i1;
          idx2 = lev1isless ? i1 : i1 - 1;
          val1 = lev1[idx1];
          val2 = lev1[idx2];
          if (lev2[i2] > val1 && lev2[i2] <= val2) break;
        }

      if (i1 == nlev1) cdoAbort("Level %g not found!", lev2[i2]);

      if (i1 - 1 == 0)  // destination levels is not covert by the first two input z levels
        {
          lev_idx1[i2] = 1;
          lev_idx2[i2] = 1;
          lev_wgt1[i2] = 0;
          lev_wgt2[i2] = (expol || IS_EQUAL(lev2[i2], val2));
        }
      else if (i1 == nlev1 - 1)  // destination level is beyond the last value of the input z field
        {
          lev_idx1[i2] = nlev1 - 2;
          lev_idx2[i2] = nlev1 - 2;
          lev_wgt1[i2] = (expol || IS_EQUAL(lev2[i2], val2));
          lev_wgt2[i2] = 0;
        }
      else  // target z values has two bounday values in input z field
        {
          lev_idx1[i2] = idx1;
          lev_idx2[i2] = idx2;
          lev_wgt1[i2] = (lev1[idx2] - lev2[i2]) / (lev1[idx2] - lev1[idx1]);
          lev_wgt2[i2] = (lev2[i2] - lev1[idx1]) / (lev1[idx2] - lev1[idx1]);
        }
      // backshift of the indices because of the two additional levels in input vertical coordinate
      lev_idx1[i2]--;
      lev_idx2[i2]--;
      /*
        printf("%d %g %d %d %g %g %d %d %g %g\n", i2, lev2[i2], idx1, idx2, lev1[idx1], lev1[idx2],
        lev_idx1[i2], lev_idx2[i2], lev_wgt1[i2], lev_wgt2[i2]);
      */
    }
}


static bool
levelDirUp(const int nlev, const double *const lev)
{
  const bool lup = (nlev > 1 && lev[1] > lev[0]);
  for (int k = 1; k < nlev - 1; ++k)
    if (lup && lev[k + 1] <= lev[k]) return false;

  return lup;
}

static bool
levelDirDown(const int nlev, const double *const lev)
{
  const bool ldown = (nlev > 1 && lev[1] < lev[0]);
  for (int k = 1; k < nlev - 1; ++k)
    if (ldown && lev[k + 1] >= lev[k]) return false;

  return ldown;
}

template <typename T>
static void
vert_gen_weights3d1d(bool expol, size_t gridsize, int nlev1, const Varray<T> &xlev1, int nlev2, const Varray<double> &lev2,
                     Varray<int> &xlev_idx1, Varray<int> &xlev_idx2, Varray<double> &xlev_wgt1, Varray<double> &xlev_wgt2)
{
  Varray<double> lev1p2(nlev1 + 2);
  Varray<double> lev_wgt1(nlev2), lev_wgt2(nlev2);
  Varray<int> lev_idx1(nlev2), lev_idx2(nlev2);

  // Check monotony of vertical levels
  for (int k = 0; k < nlev1; ++k) lev1p2[k] = xlev1[k * gridsize];
  const auto lup = levelDirUp(nlev1, lev1p2.data());
  const auto ldown = levelDirDown(nlev1, lev1p2.data());
  if (!lup && !ldown) cdoAbort("Non monotonic zaxis!");
  const double level_0 = lup ? -1.e33 : 1.e33;
  const double level_N = lup ? 1.e33 : -1.e33;

  for (size_t i = 0; i < gridsize; i++)
    {
      lev1p2[0] = level_0;
      lev1p2[nlev1 + 1] = level_N;
      for (int k = 0; k < nlev1; ++k) lev1p2[k + 1] = xlev1[k * gridsize + i];

      vert_gen_weights(expol, nlev1 + 2, lev1p2, nlev2, lev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);

      for (int k = 0; k < nlev2; ++k) xlev_idx1[k * gridsize + i] = lev_idx1[k] * gridsize + i;
      for (int k = 0; k < nlev2; ++k) xlev_idx2[k * gridsize + i] = lev_idx2[k] * gridsize + i;
      for (int k = 0; k < nlev2; ++k) xlev_wgt1[k * gridsize + i] = lev_wgt1[k];
      for (int k = 0; k < nlev2; ++k) xlev_wgt2[k * gridsize + i] = lev_wgt2[k];
    }
}

static void
vert_gen_weights3d1d(bool expol, size_t gridsize, int nlev1, Field3D &field1, int nlev2, const Varray<double> &lev2,
                     Varray<int> &lev_idx1, Varray<int> &lev_idx2, Varray<double> &lev_wgt1, Varray<double> &lev_wgt2)
{
  if (field1.memType == MemType::Float)
    {
      vert_gen_weights3d1d(expol, gridsize, nlev1, field1.vec_f, nlev2, lev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
  else
    {
      vert_gen_weights3d1d(expol, gridsize, nlev1, field1.vec_d, nlev2, lev2, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
    }
}

static bool
is_height_axis(int zaxisID)
{
  bool isHeight = false;
  if (zaxisInqType(zaxisID) == ZAXIS_REFERENCE)
    {
      char units[CDI_MAX_NAME], stdname[CDI_MAX_NAME];
      int length = CDI_MAX_NAME;
      cdiInqKeyString(zaxisID, CDI_GLOBAL, CDI_KEY_UNITS, units, &length);
      length = CDI_MAX_NAME;
      cdiInqKeyString(zaxisID, CDI_GLOBAL, CDI_KEY_STDNAME, stdname, &length);
      if (cstrIsEqual(stdname, "height") && *units == 0) isHeight = true;
    }
  return isHeight;
}

void *
Vertintgh(void *process)
{
  int nrecs;
  int varID, levelID;
  int f_heightID = -1, h_heightID = -1;;
  char stdname[CDI_MAX_NAME];
  bool extrapolate = false;

  cdoInitialize(process);

  // clang-format off
  const auto GH2HL  = cdoOperatorAdd("gh2hl",   0, 0, "height levels in meter");
  const auto GH2HLX = cdoOperatorAdd("gh2hlx",  0, 0, "height levels in meter");
  // clang-format on

  const auto operatorID = cdoOperatorID();

  if (operatorID == GH2HL)
    {
      const auto envstr = getenv("EXTRAPOLATE");
      if (envstr && isdigit((int) envstr[0]))
        {
          if (atoi(envstr) == 1) extrapolate = true;
          if (extrapolate) cdoPrint("Extrapolation of missing values enabled!");
        }
    }
  else if (operatorID == GH2HLX)
    {
      extrapolate = true;
    }

  operatorInputArg(cdoOperatorEnter(operatorID));

  int zaxisID2 = CDI_UNDEFID;
  const auto &arg1 = cdoOperatorArgv(0);
  Varray<double> heightlevs;
  if (operatorArgc() == 1 && !isdigit(arg1[0]))
    {
      auto zfp = fopen(arg1.c_str(), "r");
      if (zfp)
        {
          zaxisID2 = zaxisFromFile(zfp, arg1.c_str());
          fclose(zfp);
          if (zaxisID2 == CDI_UNDEFID) cdoAbort("Invalid zaxis description file %s!", arg1.c_str());
          const auto nlevels = zaxisInqSize(zaxisID2);
          heightlevs.resize(nlevels);
          zaxisInqLevels(zaxisID2, heightlevs.data());
        }
      else if (arg1 == "default")
        heightlevs = { 10, 50, 100, 500, 1000, 5000, 10000, 15000, 20000, 25000, 30000 };
      else
        cdoAbort("Open failed on %s", arg1.c_str());
    }
  else
    {
      heightlevs = cdoArgvToFlt(cdoGetOperArgv());
    }

  int nheightlevs = heightlevs.size();

  const auto streamID1 = cdoOpenRead(0);

  const auto vlistID1 = cdoStreamInqVlist(streamID1);
  const auto vlistID2 = vlistDuplicate(vlistID1);

  const auto taxisID1 = vlistInqTaxis(vlistID1);
  const auto taxisID2 = taxisDuplicate(taxisID1);
  vlistDefTaxis(vlistID2, taxisID2);

  const auto gridsize = vlist_check_gridsize(vlistID1);

  if (zaxisID2 == CDI_UNDEFID)
    {
      zaxisID2 = zaxisCreate(ZAXIS_HEIGHT, nheightlevs);
      zaxisDefLevels(zaxisID2, heightlevs.data());
    }

  VarList varList1, varList2;
  varListInit(varList1, vlistID1);
  varListSetUniqueMemtype(varList1);
  const auto memtype = varList1[0].memType;

  const auto nvars = vlistNvars(vlistID1);

  for (varID = 0; varID < nvars; varID++)
    {
      int length = CDI_MAX_NAME;
      cdiInqKeyString(vlistID1, varID, CDI_KEY_STDNAME, stdname, &length);
      cstrToLowerCase(stdname);

      // clang-format off
      if (cstrIsEqual(stdname, var_stdname(geometric_height_at_full_level_center))) f_heightID = varID;
      if (cstrIsEqual(stdname, var_stdname(geometric_height_at_half_level_center))) h_heightID = varID;
      // clang-format on
    }

  if (Options::cdoVerbose)
    {
      cdoPrint("Found:");
      // clang-format off
      if (-1 != f_heightID) cdoPrint("  %s -> %s", var_stdname(geometric_height_at_full_level_center), varList1[f_heightID].name);
      if (-1 != h_heightID) cdoPrint("  %s -> %s", var_stdname(geometric_height_at_half_level_center), varList1[h_heightID].name);
      // clang-format on
    }

  if (-1 == f_heightID && -1 == h_heightID) cdoAbort("%s not found!", var_stdname(geometric_height_at_full_level_center));

  const auto zaxisIDfull = (-1 != f_heightID) ? varList1[f_heightID].zaxisID : -1;
  const auto zaxisIDhalf = (-1 != h_heightID) ? varList1[h_heightID].zaxisID : -1;
  const auto nfulllevs = (-1 != zaxisIDfull) ? zaxisInqSize(zaxisIDfull) : 0;
  const auto nhalflevs = (-1 != zaxisIDhalf) ? zaxisInqSize(zaxisIDhalf) : 0;

  const auto nzaxis = vlistNzaxis(vlistID1);
  for (int index = 0; index < nzaxis; index++)
    {
      const auto zaxisID = vlistZaxis(vlistID1, index);
      const auto nlevels = zaxisInqSize(zaxisID);
      if (zaxisID == zaxisIDfull || zaxisID == zaxisIDhalf ||
          (is_height_axis(zaxisID) && (nlevels == nhalflevs || nlevels == nfulllevs)))
        vlistChangeZaxis(vlistID2, zaxisID, zaxisID2);
    }

  varListInit(varList2, vlistID2);
  varListSetMemtype(varList2, memtype);

  Varray<int> lev_idx1, lev_idx2;
  Varray<double> lev_wgt1, lev_wgt2;
  if (-1 != f_heightID)
    {
      const auto wsize = gridsize * nfulllevs;
      lev_idx1.resize(wsize);
      lev_idx2.resize(wsize);
      lev_wgt1.resize(wsize);
      lev_wgt2.resize(wsize);
    }

  Varray<int> xlev_idx1, xlev_idx2;
  Varray<double> xlev_wgt1, xlev_wgt2;
  if (-1 != h_heightID)
    {
      const auto wsize = gridsize * nhalflevs;
      xlev_idx1.resize(wsize);
      xlev_idx2.resize(wsize);
      xlev_wgt1.resize(wsize);
      xlev_wgt2.resize(wsize);
    }

  std::vector<bool> vars(nvars), varinterp(nvars);
  std::vector<std::vector<size_t>> varnmiss(nvars);
  Field3DVector vardata1(nvars), vardata2(nvars);

  const auto maxlev = std::max(std::max(nfulllevs, nhalflevs), nheightlevs);

  for (varID = 0; varID < nvars; varID++)
    {
      const auto gridID = varList1[varID].gridID;
      const auto zaxisID = varList1[varID].zaxisID;
      const auto nlevels = varList1[varID].nlevels;

      if (gridInqType(gridID) == GRID_SPECTRAL) cdoAbort("Spectral data unsupported!");

      vardata1[varID].init(varList1[varID]);

      varinterp[varID] = (zaxisID == zaxisIDfull || zaxisID == zaxisIDhalf ||
                          (is_height_axis(zaxisID) && (nlevels == nhalflevs || nlevels == nfulllevs)));

      if (varinterp[varID])
        {
          varnmiss[varID].resize(maxlev, 0);
          vardata2[varID].init(varList2[varID]);
        }
      else
        {
          if (is_height_axis(zaxisID) && nlevels > 1)
            {
              if (-1 == f_heightID && -1 != h_heightID && nlevels == (nhalflevs - 1))
                cdoAbort("%s not found (needed for %s)!", var_stdname(geometric_height_at_full_level_center), varList1[varID].name);
              else if (-1 != f_heightID && -1 == h_heightID && nlevels == (nfulllevs + 1))
                cdoAbort("%s not found (needed for %s)!", var_stdname(geometric_height_at_half_level_center), varList1[varID].name);
              else
                cdoWarning("Parameter %d has wrong number of levels, skipped! (name=%s nlevel=%d)", varID + 1, varList1[varID].name, nlevels);
            }
          varnmiss[varID].resize(nlevels);
        }
    }

  for (varID = 0; varID < nvars; ++varID)
    {
      if (varinterp[varID] && varList1[varID].timetype == TIME_CONSTANT) vlistDefVarTimetype(vlistID2, varID, TIME_VARYING);
    }

  const auto streamID2 = cdoOpenWrite(1);

  cdoDefVlist(streamID2, vlistID2);

  int tsID = 0;
  while ((nrecs = cdoStreamInqTimestep(streamID1, tsID)))
    {
      for (varID = 0; varID < nvars; ++varID)
        {
          vars[varID] = false;
          const auto nlevels = varList1[varID].nlevels;
          for (levelID = 0; levelID < nlevels; levelID++) varnmiss[varID][levelID] = 0;
        }

      taxisCopyTimestep(taxisID2, taxisID1);
      cdoDefTimestep(streamID2, tsID);

      for (int recID = 0; recID < nrecs; recID++)
        {
          cdoInqRecord(streamID1, &varID, &levelID);
          cdoReadRecord(streamID1, vardata1[varID], levelID, &varnmiss[varID][levelID]);
          vars[varID] = true;
        }

      for (varID = 0; varID < nvars; varID++)
        if (varinterp[varID]) vars[varID] = true;

      if (-1 != f_heightID && (tsID == 0 || varList1[f_heightID].timetype != TIME_CONSTANT))
        vert_gen_weights3d1d(extrapolate, gridsize, nfulllevs, vardata1[f_heightID], nheightlevs, heightlevs, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);

      if (-1 != h_heightID && (tsID == 0 || varList1[h_heightID].timetype != TIME_CONSTANT))
        vert_gen_weights3d1d(extrapolate, gridsize, nhalflevs, vardata1[h_heightID], nheightlevs, heightlevs, xlev_idx1, xlev_idx2, xlev_wgt1, xlev_wgt2);

      for (varID = 0; varID < nvars; varID++)
        {
          if (vars[varID])
            {
              if (tsID > 0 && varList1[varID].timetype == TIME_CONSTANT) continue;

              if (varinterp[varID])
                {
                  const auto nlevels = varList1[varID].nlevels;
                  if (nlevels != nfulllevs && nlevels != nhalflevs)
                    cdoAbort("Number of generalized height level differ from full/half level (param=%s)!", varList1[varID].name);

                  for (levelID = 0; levelID < nlevels; levelID++)
                    {
                      if (varnmiss[varID][levelID]) cdoAbort("Missing values unsupported for this operator!");
                    }

                  const auto missval = varList1[varID].missval;
                  if (nlevels == nfulllevs)
                    vert_interp_lev3d(gridsize, missval, vardata1[varID], vardata2[varID], nheightlevs, lev_idx1, lev_idx2, lev_wgt1, lev_wgt2);
                  else
                    vert_interp_lev3d(gridsize, missval, vardata1[varID], vardata2[varID], nheightlevs, xlev_idx1, xlev_idx2, xlev_wgt1, xlev_wgt2);

                  if (!extrapolate)
                    for (levelID = 0; levelID < nheightlevs; levelID++)
                      {
                        const auto offset = gridsize * levelID;
                        if (memtype == MemType::Float)
                          varnmiss[varID][levelID] = arrayNumMV(gridsize, &vardata2[varID].vec_f[offset], (float)missval);
                        else
                          varnmiss[varID][levelID] = arrayNumMV(gridsize, &vardata2[varID].vec_d[offset], missval);
                      }
                }

              for (levelID = 0; levelID < varList2[varID].nlevels; levelID++)
                {
                  cdoDefRecord(streamID2, varID, levelID);
                  cdoWriteRecord(streamID2, varinterp[varID] ? vardata2[varID] : vardata1[varID], levelID, varnmiss[varID][levelID]);
                }
            }
        }

      tsID++;
    }

  cdoStreamClose(streamID2);
  cdoStreamClose(streamID1);

  cdoFinish();

  return nullptr;
}
